/*----------------------------------------------------------------------------------------------------------------------------------------------------------------------------
Autor: Jan Olszak
Temat: biblioteki boost - boost::graph (BGL) - przeszukiwanie grafw algorytmem Dijkstry
----------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/


/*----------------------------------------------------------------------------------------------------------------------------------------------------------------------------
Algorytm Dijkstry przeszukiwania grafu (na podstawie Wikipedii):
0.	s - wierzchoek, od ktrego startujemy, v - dowolny inny wierzchoek

1.	Stwrz tablic d odlegosci od zrda dla wszystkich wierzchokw grafu. 
Na pocztku d[s] = 0, zas d[v] = nieskoczonosc dla wszystkich pozostaych wierzchokw.

2.	Utwrz kolejk priorytetow Q wszystkich wierzchokw grafu. 
Priorytetem kolejki jest aktualnie wyliczona odlegosc od wierzchoka zrdowego s.

3.	Dopki kolejka nie jest pusta:
3.1	Usu z kolejki wierzchoek u o najniszym priorytecie 
(wierzchoek najbliszy zrda, ktry nie zosta jeszcze rozwaony)
3.2	Dla kadego ssiada v wierzchoka u dokonaj relaksacji poprzez u: jeli d[u] + w(u,v) < d[v] 
(poprzez u da si dojsc do v szybciej ni dotychczasow ciek), to d[v]: = d[u] + w(u,v).


4.	Na kocu tablica d zawiera najkrtsze odlegoci do wszystkich wierzchokw.

Biblioteka BGL umoliwia stworzenie wizytatora, 
ktrego  metody s wywoywane kiedy odpowiednie kroki w algorytmie s wykonywane. 

DijkstraVisitor definiuje naspujcy interfejs (na podstawie dokumentacjii boost http://www.boost.org/doc/libs/1_38_0/libs/graph/doc/DijkstraVisitor.html):
u - wierzchoek, e - krawdz, g - graf

1. Initialize Vertex	vis.initialize_vertex(u, g)	void	Wywoywane raz dla kadego wierzchoka grafu, podczas jego inicjalizacji. (krok 1 algorytmu)

2. Examine Vertex		vis.examine_vertex(u, g) void	  	Wywoywane dla wierzchoka usuwanego z kolejki (krok 3.1 algorytmu). 
Zaraz potem metoda examine_edge(e,g) jest wywoywana dla wszystkich krawdzi wychodzcych z wierzchoka u

3. Examine Edge			vis.examine_edge(e, g)	void		Wywoywane dla krawdzi, kiedy jest odkrywana (krok 3.2 algorytmu)

4. Discover Vertex		vis.discover_vertex(u, g) void		Wywoywane dla wierzchoka, gdy jest odkrywany. Czyli gdy odlegosc od wierzchoka-zrda staje si skoczona. (krok 3.2 algorytmu)

5. Edge Relaxed			vis.edge_relaxed(e, g) void		Wywoywane dla krawdzi, kiedy dokonuje si przez ni relaksacji (krok 3.2 algorytmu)

6. Edge Not Relaxed		vis.edge_not_relaxed(e, g)	void	Wywoywane dla krawdzi, kiedy nie dokonuje si przez ni relaksacji (krok 3.2 algorytmu)

7. Finish Vertex		vis.finish_vertex(u, g)	void		Wywoywane dla wierzchoka po tym jak wszystkie wychodzce z niego krawdzie bd dodane do drzewa przeszukiwa, 
oraz wszystkie przylege do niego wierzchoki bd odkryte (ale zanim wychodzce z nich krawdzie bd odkryte)



Przy deklarowaniu wasnego wizytatora nie trzeba po niczym dziedziczy. Wystarczy zrobi klas, ktra ma wszystkie wyej wymienione metody.
Jest ich jednak duo i najczesciej nie bdzie potrzeby uywa wszystkich. eby uatwi tworzenie wizytatora w ktrym wywoywane s tylko niektre metody twrcy BGL stworzyli modele 
- w tym wypadku dijkstra_visitor zdefiniowany w boost/graph/dijkstra_shortest_paths.hpp oraz pomocnicz metod make_dijkstra_visitor
implementujcego DijkstraVisitor w pliku


Mona tworzy wizytatora na 2 sposoby
implementowac
make dijkstra visitator




----------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/

#include <boost/config.hpp>
#include <iostream>
#include <iterator>
#include <vector>
#include <list>
// Use boost::queue instead of std::queue because std::queue doesn't
// model Buffer; it has to top() function. -Jeremy
#include <boost/pending/queue.hpp>

#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/visitors.hpp>
#include <boost/graph/breadth_first_search.hpp>
#include <boost/graph/dijkstra_shortest_paths.hpp>
#include <boost/graph/graph_utility.hpp>


using namespace boost;


typedef adjacency_list < vecS, vecS, directedS,no_property, property < edge_weight_t, int > >  DirectedGraph;
typedef graph_traits<DirectedGraph>::edge_descriptor EdgeDescriptor;
typedef graph_traits<DirectedGraph>::vertex_descriptor VertexDescriptor;
typedef std::pair<int, int> Edge;


// Wierzchoki bd typu int, N - liczba wierzchokw
enum { A, B, C, N };

// Pomocna tablica do wypisywania na ekran
char name[] = "ABCDE";

class MyVisitor {
public:
	MyVisitor() { }
	void initialize_vertex(VertexDescriptor u,DirectedGraph g)
	{
		std::cout << "initialize_vertex " << u << std::endl;
	}
	void discover_vertex(VertexDescriptor u,DirectedGraph g){
		std::cout << "discover_vertex "<< u << std::endl;

	}
	void examine_vertex(VertexDescriptor u,DirectedGraph g){
		std::cout << "examine_vertex "<< u << std::endl;

	}
	void examine_edge(EdgeDescriptor e,DirectedGraph g){
		std::cout << "examine_edge "<< e << std::endl;

	}
	void edge_relaxed(EdgeDescriptor e,DirectedGraph g){
		std::cout << "edge_relaxed "<< e <<std::endl;

	}
	void edge_not_relaxed(EdgeDescriptor e,DirectedGraph g){
		std::cout << "edge_not_relaxed "<< e <<std::endl;

	}
	void finish_vertex(VertexDescriptor u,DirectedGraph g){
		std::cout << "finish_vertex "<< u << std::endl;

	}



};

void print(std::string title,DirectedGraph g, std::vector<VertexDescriptor>  &p, std::vector<int> &d ){
	// Pomocnicza funkcja wypisujca.


	std::cout << title << std::endl;
	graph_traits < DirectedGraph >::vertex_iterator vi, vend;
	for (tie(vi, vend) = vertices(g); vi != vend; ++vi) {
		std::cout << "najmniejsza odleglosc(" << name[*vi] << ") = " << d[*vi] << ", ";
		std::cout << "rodzic(" << name[*vi] << ") = " << name[p[*vi]] << std::
			endl;
	}
	std::cout << std::endl;
}




void firstExample(){
	/*
	Przykad, w ktrym uywamy wasnej klasy-wizytatora, 
	w ktrej trzeba definiowa wszystkie metody interfejsu DijkstraVisitor
	*/

	// Tablica krawdzi
	Edge edge_array[] = {Edge(A,B), Edge(A,C), Edge(B,C)};

	// Tablica wag dla krawdzi
	int weights[] = { 1, 9, 3};

	// Liczba krawdzi
	int num_arcs = sizeof(edge_array) / sizeof(Edge);

	// Tworzymy graf g
	DirectedGraph g(edge_array, edge_array + num_arcs, weights, N);

	// Kontenery, do ktrych dijkstra_shortest_paths() wypisze wynik
	std::vector<VertexDescriptor> p(num_vertices(g));
	std::vector<int> d(num_vertices(g));
	VertexDescriptor s = vertex(A, g);

	// Nasz wizytator, wszystkie metody musz by zadeklarowane
	MyVisitor myVisitor;

	// Wywoanie algorymu Dijkstry. Argumenty podane przy pomocy bgl_named_params. http://www.boost.org/doc/libs/1_38_0/libs/graph/doc/bgl_named_params.html
	dijkstra_shortest_paths(g, s, predecessor_map(&p[0]).distance_map(&d[0]).visitor(myVisitor)    );

	print("Pierwszy przyklad: ",g,p,d);

}




template <class NewGraph, class Tag>
struct MyPartVisitor  
	: public base_visitor<MyPartVisitor<NewGraph, Tag> >
{
	typedef Tag event_filter;

	MyPartVisitor(NewGraph& graph) : new_g(graph) { }

	template <class Edge, class Graph>
	void operator()(Edge e, Graph& g) {
		add_edge(source(e, g), target(e, g), new_g);
		std::cout << "kopiuje krawedz"<< e << std::endl;
	}
private:
	NewGraph& new_g;

};


template <class NewGraph, class Tag>
inline MyPartVisitor<NewGraph, Tag>
makeMyPartVisitor(NewGraph& g, Tag) {
	return MyPartVisitor<NewGraph, Tag>(g); // funkcje moga dedukowac typy argumentow, ktore sie do nich przekazuje
}

void secondExample(){
	/*
	Przykad, w ktrym uywamy wasnego wizytatora, ktory implementuje tylko jedna metode interfejsu.
	W tym wypadku, chcemy przekopiowac graf g do grafu g_copy
	*/

	// Tablica krawdzi
	Edge edge_array[] = {Edge(A,B), Edge(A,C), Edge(B,C)};

	// Tablica wag dla krawdzi
	int weights[] = { 1, 9, 3};

	// Liczba krawdzi
	int num_arcs = sizeof(edge_array) / sizeof(Edge);

	// Tworzymy graf g
	DirectedGraph g(edge_array, edge_array + num_arcs, weights, N);

	// Tworzymy graf, do ktrego przekopiujemy graf g
	DirectedGraph g_copy(edge_array, edge_array + num_arcs, weights, N);

	// Kontenery, do ktrych dijkstra_shortest_paths() wypisze wynik
	std::vector<VertexDescriptor> p(num_vertices(g));
	std::vector<int> d(num_vertices(g));

	// Wierzchoek pocztkowy algorytmu - source
	VertexDescriptor s = vertex(A, g);


	// Wywoanie algorymu Dijkstry. Argumenty podane przy pomocy bgl_named_params. http://www.boost.org/doc/libs/1_38_0/libs/graph/doc/bgl_named_params.html

	dijkstra_shortest_paths(g, s, predecessor_map(&p[0]).distance_map(&d[0]).visitor(make_dijkstra_visitor(makeMyPartVisitor(g_copy, on_examine_edge())))    );

	print("Drugi przyklad, skopiowany graf:",g_copy,p,d);

}


int main(int, char *[])
{
	firstExample();
	secondExample();

	std::system("pause");

	return EXIT_SUCCESS;

}